@@ -30,7 +30,7 @@ from calibre.utils.filenames import (
     is_case_sensitive, samefile, hardlink_file, ascii_filename,
     WindowsAtomicFolderMove, atomic_rename, remove_dir_if_empty,
     copytree_using_links, copyfile_using_links)
-from calibre.utils.magick.draw import save_cover_data_to
+from calibre.utils.img import save_cover_data_to
 from calibre.utils.formatter_functions import load_user_template_functions
 from calibre.db.tables import (OneToOneTable, ManyToOneTable, ManyToManyTable,
         SizeTable, FormatsTable, AuthorsTable, IdentifiersTable, PathTable,
@@ -47,6 +47,7 @@ Differences in semantics from pysqlite:
 '''
 CUSTOM_DATA_TYPES = frozenset(['rating', 'text', 'comments', 'datetime',
     'int', 'float', 'bool', 'series', 'composite', 'enumeration'])
+WINDOWS_RESERVED_NAMES = frozenset('CON PRN AUX NUL COM1 COM2 COM3 COM4 COM5 COM6 COM7 COM8 COM9 LPT1 LPT2 LPT3 LPT4 LPT5 LPT6 LPT7 LPT8 LPT9'.split())
 
 
 class DynamicFilter(object):  # {{{
@@ -1139,6 +1140,8 @@ class DB(object):
         if not author:
             author = ascii_filename(_('Unknown')).decode(
                     'ascii', 'replace')
+        if author.upper() in WINDOWS_RESERVED_NAMES:
+            author += 'w'
         return '%s/%s%s' % (author, title, book_id)
 
     def construct_file_name(self, book_id, title, author, extlen):
@@ -1295,7 +1298,7 @@ class DB(object):
         except EnvironmentError:
             pass  # Cover doesn't exist
 
-    def copy_cover_to(self, path, dest, windows_atomic_move=None, use_hardlink=False):
+    def copy_cover_to(self, path, dest, windows_atomic_move=None, use_hardlink=False, report_file_size=None):
         path = os.path.abspath(os.path.join(self.library_path, path, 'cover.jpg'))
         if windows_atomic_move is not None:
             if not isinstance(dest, basestring):
@@ -1318,6 +1321,10 @@ class DB(object):
 
                 with f:
                     if hasattr(dest, 'write'):
+                        if report_file_size is not None:
+                            f.seek(0, os.SEEK_END)
+                            report_file_size(f.tell())
+                            f.seek(0)
                         shutil.copyfileobj(f, dest)
                         if hasattr(dest, 'flush'):
                             dest.flush()
@@ -1350,7 +1357,7 @@ class DB(object):
         with f:
             return True, f.read(), stat.st_mtime
 
-    def set_cover(self, book_id, path, data):
+    def set_cover(self, book_id, path, data, no_processing=False):
         path = os.path.abspath(os.path.join(self.library_path, path))
         if not os.path.exists(path):
             os.makedirs(path)
@@ -1368,14 +1375,18 @@ class DB(object):
                     time.sleep(0.2)
                     os.remove(path)
         else:
-            try:
-                save_cover_data_to(data, path)
-            except (IOError, OSError):
-                time.sleep(0.2)
-                save_cover_data_to(data, path)
+            if no_processing:
+                with lopen(path, 'wb') as f:
+                    f.write(data)
+            else:
+                try:
+                    save_cover_data_to(data, path)
+                except (IOError, OSError):
+                    time.sleep(0.2)
+                    save_cover_data_to(data, path)
 
     def copy_format_to(self, book_id, fmt, fname, path, dest,
-                       windows_atomic_move=None, use_hardlink=False):
+                       windows_atomic_move=None, use_hardlink=False, report_file_size=None):
         path = self.format_abspath(book_id, fmt, fname, path)
         if path is None:
             return False
@@ -1396,6 +1407,10 @@ class DB(object):
         else:
             if hasattr(dest, 'write'):
                 with lopen(path, 'rb') as f:
+                    if report_file_size is not None:
+                        f.seek(0, os.SEEK_END)
+                        report_file_size(f.tell())
+                        f.seek(0)
                     shutil.copyfileobj(f, dest)
                 if hasattr(dest, 'flush'):
                     dest.flush()
@@ -1434,7 +1449,7 @@ class DB(object):
                         if wam is not None:
                             wam.close_handles()
 
-    def add_format(self, book_id, fmt, stream, title, author, path, current_name):
+    def add_format(self, book_id, fmt, stream, title, author, path, current_name, mtime=None):
         fmt = ('.' + fmt.lower()) if fmt else ''
         fname = self.construct_file_name(book_id, title, author, len(fmt))
         path = os.path.join(self.library_path, path)
@@ -1463,8 +1478,12 @@ class DB(object):
             with lopen(dest, 'wb') as f:
                 shutil.copyfileobj(stream, f)
                 size = f.tell()
+            if mtime is not None:
+                os.utime(dest, (mtime, mtime))
         elif os.path.exists(dest):
             size = os.path.getsize(dest)
+            if mtime is not None:
+                os.utime(dest, (mtime, mtime))
 
         return size, fname
 
@@ -1723,4 +1742,17 @@ class DB(object):
         self.execute('UPDATE books SET path=? WHERE id=?', (path.replace(os.sep, '/'), book_id))
         vals = [(book_id, fmt, size, name) for fmt, size, name in formats]
         self.executemany('INSERT INTO data (book,format,uncompressed_size,name) VALUES (?,?,?,?)', vals)
+
+    def backup_database(self, path):
+        # We have to open a new connection to self.dbpath, until this issue is fixed:
+        # https://github.com/rogerbinns/apsw/issues/199
+        dest_db = apsw.Connection(path)
+        source = apsw.Connection(self.dbpath)
+        with dest_db.backup('main', source, 'main') as b:
+            while not b.done:
+                b.step(100)
+        source.close()
+        dest_db.cursor().execute('DELETE FROM metadata_dirtied; VACUUM;')
+        dest_db.close()
+
     # }}}
