@@ -4,21 +4,27 @@
 
 from __future__ import (unicode_literals, division, absolute_import,
                         print_function)
-import sys, os, json
+import sys, os, json, re
 from base64 import standard_b64encode, standard_b64decode
-from collections import defaultdict
+from collections import defaultdict, OrderedDict
 from itertools import count
 from functools import partial
 from future_builtins import map
 from urlparse import urlparse
+from urllib import quote
 
 from cssutils import replaceUrls
+from cssutils.css import CSSRule
 
+from calibre import prepare_string_for_xml
+from calibre.ebooks import parse_css_length
 from calibre.ebooks.oeb.base import (
-    OEB_DOCS, OEB_STYLES, rewrite_links, XPath, urlunquote, XLINK, XHTML_NS, OPF, XHTML)
+    OEB_DOCS, OEB_STYLES, rewrite_links, XPath, urlunquote, XLINK, XHTML_NS, OPF, XHTML, EPUB_NS)
 from calibre.ebooks.oeb.iterator.book import extract_book
 from calibre.ebooks.oeb.polish.container import Container as ContainerBase
 from calibre.ebooks.oeb.polish.cover import set_epub_cover, find_cover_image
+from calibre.ebooks.oeb.polish.css import transform_css
+from calibre.ebooks.css_transform_rules import StyleDeclaration
 from calibre.ebooks.oeb.polish.toc import get_toc
 from calibre.ebooks.oeb.polish.utils import guess_type
 from calibre.utils.short_uuid import uuid4
@@ -42,7 +48,73 @@ def encode_url(name, frag=''):
 
 def decode_url(x):
     parts = x.split('#', 1)
-    return decode_component(parts[0]), parts[1] or ''
+    return decode_component(parts[0]), (parts[1] if len(parts) > 1 else '')
+
+absolute_units = frozenset('px mm cm pt in pc q'.split())
+length_factors = {'mm':2.8346456693, 'cm':28.346456693, 'in': 72, 'pc': 12, 'q':0.708661417325}
+
+def convert_fontsize(length, unit, base_font_size=16.0, dpi=96.0):
+    ' Convert font size to rem so that font size scaling works. Assumes the document has the specified base font size in px '
+    if unit == 'px':
+        return length/base_font_size
+    pt_to_px = dpi / 72.0
+    pt_to_rem = pt_to_px / base_font_size
+    return length * length_factors.get(unit, 1) * pt_to_rem
+
+def transform_declaration(decl):
+    decl = StyleDeclaration(decl)
+    changed = False
+    for prop, parent_prop in tuple(decl):
+        if prop.name in {'page-break-before', 'page-break-after', 'page-break-inside'}:
+            changed = True
+            name = prop.name.partition('-')[2]
+            for prefix in ('', '-webkit-column-'):
+                # Note that Firefox does not support break-after at all
+                # https://bugzil.la/549114
+                decl.set_property(prefix + name, prop.value, prop.priority)
+            decl.remove_property(prop, parent_prop)
+        elif prop.name == 'font-size':
+            l, unit = parse_css_length(prop.value)
+            if unit in absolute_units:
+                changed = True
+                l = convert_fontsize(l, unit)
+                decl.change_property(prop, parent_prop, str(l) + 'rem')
+    return changed
+
+def transform_sheet(sheet):
+    changed = False
+    for rule in sheet.cssRules.rulesOfType(CSSRule.STYLE_RULE):
+        if transform_declaration(rule.style):
+            changed = True
+    return changed
+
+def check_for_maths(root):
+    for x in root.iterdescendants('{*}math'):
+        return True
+    for s in root.iterdescendants(XHTML('script')):
+        if s.get('type') == 'text/x-mathjax-config':
+            return True
+    return False
+
+def get_length(root):
+    strip_space = re.compile(r'\s+')
+    ans = 0
+    def count(elem):
+        num = 0
+        tname = elem.tag.rpartition('}')[-1].lower()
+        if elem.text and tname not in 'script style':
+            num += len(strip_space.sub(elem.text, ''))
+        if elem.tail:
+            num += len(strip_space.sub(elem.tail, ''))
+        if tname in 'img svg':
+            num += 2000
+        return num
+
+    for body in root.iterdescendants(XHTML('body')):
+        ans += count(body)
+        for elem in body.iterdescendants('*'):
+            ans += count(elem)
+    return ans
 
 class Container(ContainerBase):
 
@@ -68,20 +140,34 @@ class Container(ContainerBase):
             'is_comic': input_fmt.lower() in {'cbc', 'cbz', 'cbr', 'cb7'},
             'raster_cover_name': raster_cover_name,
             'title_page_name': titlepage_name,
+            'has_maths': False,
+            'total_length': 0,
+            'spine_length': 0,
         }
         # Mark the spine as dirty since we have to ensure it is normalized
         for name in data['spine']:
             self.parsed(name), self.dirty(name)
+        self.transform_css()
         self.virtualized_names = set()
         self.virtualize_resources()
         def manifest_data(name):
             mt = (self.mime_map.get(name) or 'application/octet-stream').lower()
-            return {
+            ans = {
                 'size':os.path.getsize(self.name_path_map[name]),
                 'is_virtualized': name in self.virtualized_names,
                 'mimetype':mt,
-                'is_html': mt in OEB_DOCS
+                'is_html': mt in OEB_DOCS,
             }
+            if ans['is_html']:
+                root = self.parsed(name)
+                ans['length'] = l = get_length(root)
+                self.book_render_data['total_length'] += l
+                if name in data['spine']:
+                    self.book_render_data['spine_length'] += l
+                ans['has_maths'] = hm = check_for_maths(root)
+                if hm:
+                    self.book_render_data['has_maths'] = True
+            return ans
         data['files'] = {name:manifest_data(name) for name in set(self.name_path_map) - excluded_names}
         self.commit()
         for name in excluded_names:
@@ -90,12 +176,18 @@ class Container(ContainerBase):
             f.write(json.dumps(self.book_render_data, ensure_ascii=False).encode('utf-8'))
 
     def create_cover_page(self, input_fmt):
+        templ = '''
+        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
+        <head><style>
+        html, body, img { height: 100%%; display: block; margin: 0; padding: 0; border-width: 0; }
+        img { width: auto; margin-left:auto; margin-right: auto; }
+        </style></head><body><img src="%s"/></body></html>
+        '''
         if input_fmt == 'epub':
             def cover_path(action, data):
                 if action == 'write_image':
                     data.write(BLANK_JPEG)
-            return set_epub_cover(self, cover_path, (lambda *a: None))
-        from calibre.ebooks.oeb.transforms.cover import CoverManager
+            return set_epub_cover(self, cover_path, (lambda *a: None), options={'template':templ})
         raster_cover_name = find_cover_image(self, strict=True)
         if raster_cover_name is None:
             item = self.generate_item(name='cover.jpeg', id_prefix='cover')
@@ -104,8 +196,7 @@ class Container(ContainerBase):
             dest.write(BLANK_JPEG)
         item = self.generate_item(name='titlepage.html', id_prefix='titlepage')
         titlepage_name = self.href_to_name(item.get('href'), self.opf_name)
-        templ = CoverManager.SVG_TEMPLATE
-        raw = templ % self.name_to_href(raster_cover_name, titlepage_name)
+        raw = templ % prepare_string_for_xml(self.name_to_href(raster_cover_name, titlepage_name), True)
         with self.open(titlepage_name, 'wb') as f:
             f.write(raw.encode('utf-8'))
         spine = self.opf_xpath('//opf:spine')[0]
@@ -114,6 +205,9 @@ class Container(ContainerBase):
         self.dirty(self.opf_name)
         return raster_cover_name, titlepage_name
 
+    def transform_css(self):
+        transform_css(self, transform_sheet=transform_sheet, transform_style=transform_declaration)
+
     def virtualize_resources(self):
 
         changed = set()
@@ -139,8 +233,11 @@ class Container(ContainerBase):
             url, frag = purl.path, purl.fragment
             name = self.href_to_name(url, base)
             if name:
-                frag = urlunquote(frag)
-                url = resource_template.format(encode_url(name, frag))
+                if self.has_name(name):
+                    frag = urlunquote(frag)
+                    url = resource_template.format(encode_url(name, frag))
+                else:
+                    url = 'missing:' + quote(name)
                 changed.add(base)
             return url
 
@@ -161,6 +258,7 @@ class Container(ContainerBase):
                         a.set('data-' + link_uid, json.dumps({'name':parts[0], 'frag':parts[1]}, ensure_ascii=False))
                     else:
                         a.set('target', '_blank')
+                        a.set('rel', 'noopener noreferrer')
                     changed.add(name)
             elif mt == 'image/svg+xml':
                 self.virtualized_names.add(name)
@@ -186,10 +284,45 @@ def split_name(name):
 
 boolean_attributes = frozenset('allowfullscreen,async,autofocus,autoplay,checked,compact,controls,declare,default,defaultchecked,defaultmuted,defaultselected,defer,disabled,enabled,formnovalidate,hidden,indeterminate,inert,ismap,itemscope,loop,multiple,muted,nohref,noresize,noshade,novalidate,nowrap,open,pauseonexit,readonly,required,reversed,scoped,seamless,selected,sortable,truespeed,typemustmatch,visible'.split(','))  # noqa
 
+EPUB_TYPE_MAP = {k:'doc-' + k for k in (
+    'abstract acknowledgements afterword appendix biblioentry bibliography biblioref chapter colophon conclusion cover credit'
+    ' credits dedication epigraph epilogue errata footnote footnotes forward glossary glossref index introduction noteref notice'
+    ' pagebreak pagelist part preface prologue pullquote qna locator subtitle title toc').split(' ')}
+for k in 'figure term definition directory list list-item table row cell'.split(' '):
+    EPUB_TYPE_MAP[k] = k
+
+EPUB_TYPE_MAP['help'] = 'doc-tip'
+
+def map_epub_type(epub_type, attribs, elem):
+    val = EPUB_TYPE_MAP.get(epub_type.lower())
+    if val:
+        role = None
+        in_attribs = None
+        for i, x in enumerate(attribs):
+            if x[0] == 'role':
+                role = x[1]
+                in_attribs = i
+                break
+        else:
+            role = elem.get('role')
+        roles = OrderedDict([(k, True) for k in role.split()]) if role else OrderedDict()
+        if val not in roles:
+            roles[val] = True
+        role = ' '.join(roles.iterkeys())
+        if in_attribs is None:
+            attribs.append(['role', role])
+        else:
+            attribs[i] = ['role', role]
+
 def serialize_elem(elem, nsmap):
     ns, name = split_name(elem.tag)
+    nl = name.lower()
+    if ns == EPUB_NS:
+        ns, name = None, 'epub-' + name
+    if nl == 'meta':
+        return  # Filter out <meta> tags as they have unknown side-effects
     if name.lower() in {'img', 'script', 'link', 'image', 'style'}:
-        name = name.lower()
+        name = nl
     ans = {'n':name}
     if elem.text:
         ans['x'] = elem.text
@@ -207,6 +340,9 @@ def serialize_elem(elem, nsmap):
             if val and val.lower() in (al, ''):
                 attribs.append([al, al])
             continue
+        if attr_ns == EPUB_NS and al == 'type':
+            map_epub_type(val, attribs, elem)
+            continue
         attrib = [aname, val]
         if attr_ns:
             attr_ns = nsmap[attr_ns]
@@ -260,10 +396,11 @@ def html_as_dict(root):
         elem, node = stack.pop()
         for i, child in enumerate(elem.iterchildren('*')):
             cnode = serialize_elem(child, nsmap)
-            tags.append(cnode)
-            child_tree_node = [len(tags)-1]
-            node.append(child_tree_node)
-            stack.append((child, child_tree_node))
+            if cnode is not None:
+                tags.append(cnode)
+                child_tree_node = [len(tags)-1]
+                node.append(child_tree_node)
+                stack.append((child, child_tree_node))
     ns_map = [ns for ns, nsnum in sorted(nsmap.iteritems(), key=lambda x: x[1])]
     return {'ns_map':ns_map, 'tag_map':tags, 'tree':tree}
 
@@ -271,4 +408,4 @@ def render(pathtoebook, output_dir, book_hash=None):
     Container(pathtoebook, output_dir, book_hash=book_hash)
 
 if __name__ == '__main__':
-    c = Container(sys.argv[-2], sys.argv[-1])
+    render(sys.argv[-2], sys.argv[-1])
